import time
import math

from typing import Any, Dict, Sequence

from coba.utilities import PackageChecker
from coba.simulations import Context, Action
from coba.encodings import InteractionTermsEncoder
from coba.learners.core import Learner, Info

class RegCBLearner(Learner):
    """A learner using the RegCB algorithm by Foster et al.
        and the online bin search implementation by Bietti et al. 

    References:
        Foster, Dylan, Alekh Agarwal, Miroslav Dudík, Haipeng Luo, and Robert Schapire.
        "Practical contextual bandits with regression oracles." In International 
        Conference on Machine Learning, pp. 1539-1548. PMLR, 2018.

        Bietti, Alberto, Alekh Agarwal, and John Langford.
        "A contextual bandit bake-off." arXiv preprint 
        arXiv:1802.04064 (2018).
    """

    @property
    def params(self) -> Dict[str, Any]:
        """The parameters of the learner.

        See the base class for more information
        """
        dict = {'family': 'RegCB', 'beta': self._beta, 'alpha': self._alpha, 'interactions': self._interactions}
        return dict

    def __init__(self, *, beta: float, alpha: float, learning_rate:float=0.1, interactions: Sequence[str] = ['a', 'ax']) -> None:
        """Instantiate a RegCBLearner.

        Args:
            beta : square-loss tolerance
            alpha: confidence bounds precision
            interactions: the set of interactions the learner will use. x refers to context and a refers to actions, 
                e.g. xaa would mean interactions between context, actions and actions. 
        """

        PackageChecker.sklearn("RegCBLearner")
        from sklearn.feature_extraction import FeatureHasher

        self._beta  = beta
        self._alpha = alpha
        self._iter  = 0

        self._core_model = []

        self._interactions = interactions
        self._interactions_encoder = InteractionTermsEncoder(interactions)

        self._times         = [0,0,0,0]
        self._learning_rate = learning_rate

        self._feature_hasher = FeatureHasher(input_type='pair')

    def predict(self, context: Context, actions: Sequence[Action]) -> Sequence[float]:
        """Determine a PMF with which to select the given actions.

        Args:
            context: The context we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The probability of taking each action. See the base class for more information.
        """

        import numpy as np
        from scipy import sparse

        if self._iter == 0:
            if isinstance(context,dict) or isinstance(actions[0],dict):
                self._core_model = sparse.csr_matrix(self._featurize(context, actions[0]).shape)
            else:
                self._core_model = np.zeros(self._featurize(context, actions[0]).shape)

        if self._iter == 200:
            self._times = [0,0,0,0]

        if (self._iter < 200):
            return [1/len(actions)] * len(actions)

        else:
            maxScore  = -float('inf')
            maxAction = None

            for action in actions:
                features = self._featurize(context,action)
                score = self._bin_search(features, len(actions))

                if score > maxScore:
                    maxAction = action
                    maxScore  = score

            return [int(action == maxAction) for action in actions]

    def learn(self, context: Context, action: Action, reward: float, probability: float, info: Info) -> None:
        """Learn from the given interaction.

        Args:
            context: The context we're learning about. See the base class for more information.
            action: The action that was selected in the context. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
            probability: The probability that the given action was taken.
            info: Optional information provided during prediction step for use in learning.
        """

        assert 0 <= reward and reward <= 1, "This regCB implementation assumes reward is in [0,1]."

        start = time.time()
        features = self._featurize(context, action)
        self._core_model = self._update_model(self._core_model, features, reward, 1)
        self._times[2] += time.time()-start

        self._iter += 1

        # if (self._iter-200) % 200 == 0 and self._iter > 200:
        #     print(f'avg phi time: {round(self._times[0]/(self._iter-200),2)}')
        #     print(f'avg bin time: {round(self._times[1]/(self._iter-200),2)}')
        #     print(f'avg lrn time: {round(self._times[2]/(self._iter-200),2)}')

    def _bin_search(self, features, K_t) -> float:

        start = time.time()

        y_u = 2
        w   = 1

        f_u_a_w = self._update_model(self._core_model, features, y_u, w)
        f_x_t_a = self._predict_model(self._core_model, features)
        s_u_a   = (self._predict_model(f_u_a_w, features) - f_x_t_a) / w

        obj = lambda w: w*(f_x_t_a-y_u)**2 - w*(f_x_t_a+s_u_a*w-y_u)**2

        lower_search_bound = 0
        upper_search_bound = (f_x_t_a-y_u)/(-s_u_a)
        width_search_bound = upper_search_bound - lower_search_bound

        constraint = self._alpha * math.log(K_t)

        w_old = lower_search_bound
        w_now = lower_search_bound + 1/2*width_search_bound
        o     = obj(w_now)

        while abs(w_now-w_old) > width_search_bound*(1/2)**30 or o >= constraint:
            w_diff = abs(w_now-w_old)
            w_old  = w_now
            if o < constraint:
                w_now += w_diff/2
            else:
                w_now -= w_diff/2
            o = obj(w_now)

        self._times[1] += time.time() - start

        return f_x_t_a + s_u_a*w_now

    def _predict_model(self, model, features):
        import numpy as np
        import scipy.sparse as sp

        if sp.issparse(model):
            return model.multiply(features).data.sum()
        else:
            return np.dot(model, features)

    def _update_model(self, model, features, value, importance):
        error = self._predict_model(model, features) - value
        return model - self._learning_rate*features*error*importance

    def _featurize(self, context, action) -> None:
        import numpy as np

        features = self._interactions_encoder.encode(x=context,a=action)

        if isinstance(features[0],tuple):
            return self._feature_hasher.fit_transform([features])[0]
        else:
            return np.array(features)



from typing import Sequence, Iterable, Dict, Any

from coba.simulations.core import Simulation, Interaction
from coba.simulations.filters import SimulationFilter

class SimSourceFilters(Simulation):

    def __init__(self, source: Simulation, filters: Sequence[SimulationFilter]):
        
        if isinstance(source, SimSourceFilters):
            self._source  = source._source
            self._filters = list(source._filters) + list(filters)
        else:
            self._source  = source
            self._filters = list(filters)

    @property
    def params(self) -> Dict[str, Any]:
        params = self._safe_params(self._source)

        for filter in self._filters:
            params.update(self._safe_params(filter))

        return params

    @property
    def source_repr(self) -> str:

        source_params = self._safe_params(self._source)

        return str(source_params) if source_params else self._source.__class__.__name__

    def read(self) -> Iterable[Interaction]:
        interactions = self._source.read()

        for filter in self._filters:
            interactions = filter.filter(interactions)

        return interactions

    def _safe_params(self, obj) -> Dict[str, Any]:
        try:
            return obj.params
        except AttributeError:
            return {}



import collections

from statistics import mean, median, stdev
from numbers import Number
from abc import abstractmethod, ABC
from itertools import islice, chain
from typing import Hashable, Optional, Sequence, cast, Union, Iterable, Dict, Any, List, Tuple

from coba.random import CobaRandom
from coba.pipes import Filter
from coba.statistics import iqr

from coba.simulations.core import Interaction

class SimulationFilter(Filter[Iterable[Interaction],Iterable[Interaction]], ABC):

    @property
    @abstractmethod
    def params(self) -> Dict[str, Any]:
        ...

    @abstractmethod
    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        """Apply a filter to a Simulation's interactions."""
        ...

class Identity(SimulationFilter):

    @property
    def params(self) -> Dict[str, Any]:
        return { }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        return interactions

class Shuffle(SimulationFilter):
    
    def __init__(self, seed:Optional[int]) -> None:
        
        if seed is not None and (not isinstance(seed,int) or seed < 0):
            raise ValueError(f"Invalid parameter for Shuffle: {seed}. An optional integer value >= 0 was expected.")

        self._seed = seed

    @property
    def params(self) -> Dict[str, Any]:
        return { "shuffle": self._seed }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]: 
        return CobaRandom(self._seed).shuffle(list(interactions))

    def __repr__(self) -> str:
        return str(self.params)

class Take(SimulationFilter):
    
    def __init__(self, count:Optional[int]) -> None:
        
        if count is not None and (not isinstance(count,int) or count < 0):
            raise ValueError(f"Invalid parameter for Take: {count}. An optional integer value >= 0 was expected.")

        self._count = count

    @property
    def params(self) -> Dict[str, Any]:
        return { "take": self._count }    

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        if self._count is None: return interactions

        materialized = list(islice(interactions,self._count))

        return materialized if len(materialized) == self._count else []

    def __repr__(self) -> str:
        return str(self.params)

class Sort(SimulationFilter):

    def __init__(self, *indexes: Union[int,Sequence[int]]) -> None:
        
        flat_indexes = cast(Sequence[int], indexes[0] if isinstance(indexes[0], collections.Sequence) else indexes)

        if not isinstance(flat_indexes, collections.Sequence) or not isinstance(flat_indexes[0],int):
            raise ValueError(f"Invalid parameter for Sort: {flat_indexes}. A sequence of integers was expected.")

        self._indexes = flat_indexes

    @property
    def params(self) -> Dict[str, Any]:
        return { "sort": self._indexes }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
        
        return sorted(interactions, key=lambda interaction: tuple(interaction.context[i] for i in self._indexes))

    def __repr__(self) -> str:
        return str(self.params)

class Scale(SimulationFilter):

    def __init__(self, shift:Union[Number,str] ="min", scale:Union[Number,str]="minmax", using:Optional[int]=None):
        assert isinstance(shift,Number) or shift in ["min","mean","med"]
        assert isinstance(scale,Number) or scale in ["minmax","std","iqr"]

        self._shift = shift
        self._scale = scale
        self._using = using

    @property
    def params(self) -> Dict[str, Any]:
        return { "scale_shift": self._shift, "scale_scale":self._scale, "scale_using":self._using }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        underlying_iterable = iter(interactions)
        using_interactions  = list(islice(underlying_iterable,self._using))

        shifts  : Dict[Hashable,float]     = collections.defaultdict(lambda:0)
        scales  : Dict[Hashable,float]     = collections.defaultdict(lambda:1)
        features: Dict[Hashable,List[Any]] = collections.defaultdict(list)

        for interaction in using_interactions:            
            for key,value in self._context_to_key_values(interaction.context):
                features[key].append(value)

        for feature_key,feature_values in features.items():

            if isinstance(feature_values[0],str):
                continue 

            if isinstance(self._shift, Number):
                shifts[feature_key] = self._shift

            if self._shift == "min":
                shifts[feature_key] = min(feature_values)

            if self._shift == "mean":
                shifts[feature_key] = mean(feature_values)

            if self._shift == "med":
                shifts[feature_key] = median(feature_values)

            if isinstance(self._scale, Number):
                num = self._scale
                den = 1
   
            if self._scale == "std":
                num = 1
                den = stdev(feature_values)

            if self._scale == "minmax":
                num = 1
                den = max(feature_values)-min(feature_values)

            if self._scale == "iqr":
                num = 1
                den = iqr(feature_values)

            scales[feature_key] = num/den if den != 0 else 1

        for interaction in chain.from_iterable([using_interactions, underlying_iterable]):

            kv_scaled_context = {}

            for key,value in self._context_to_key_values(interaction.context):
                if isinstance(value,Number):
                    kv_scaled_context[key] = (value-shifts[key])*scales[key]
                else:
                    kv_scaled_context[key] = value

            if interaction.context is None:
                final_context = None
            elif isinstance(interaction.context,dict):
                final_context = kv_scaled_context
            elif isinstance(interaction.context,tuple):
                final_context = tuple( kv_scaled_context[k] for k,_ in self._context_to_key_values(interaction.context))
            else:
                final_context = kv_scaled_context[1]

            yield Interaction(final_context, interaction.actions, reveals=interaction.reveals, **interaction.results)

    def _context_to_key_values(self,context) -> Sequence[Tuple[Hashable,Any]]:
        
        if isinstance(context,dict ): return context.items()
        if isinstance(context,tuple): return enumerate(context)
        if context is not None      : return [(1,context)]

        return []

    def __repr__(self) -> str:
        return str(self.params)

class CycleNoise(SimulationFilter):

    def __init__(self, after:int=0):
        self._after = after

    @property
    def params(self) -> Dict[str, Any]:
        return { "cycle_after": self._after }

    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:

        underlying_iterable     = iter(interactions)
        sans_noise_interactions = islice(underlying_iterable, self._after)
        with_noise_interactions = underlying_iterable

        for interaction in sans_noise_interactions:
            yield interaction

        for interaction in with_noise_interactions:
            
            direct_reveals = list(interaction.reveals)
            cycled_reveals  = direct_reveals[1:]+direct_reveals[:1]

            context = interaction.context
            actions = interaction.actions
            results = {**interaction.results, "direct_reveals":direct_reveals}

            yield Interaction(context, actions, reveals=cycled_reveals, **results)

    def __repr__(self) -> str:
        return str(self.params)


from coba.utilities import HashableDict
import collections

from abc import abstractmethod, ABC
from numbers import Number
from itertools import repeat, chain
from typing import Optional, Sequence, List, Callable, Hashable, Any, Union, Iterable, cast, overload, Dict

from coba.random import CobaRandom

from coba.pipes import (
    Pipe, Source, Filter,
    CsvReader, ArffReader, LibSvmReader, ManikReader, 
    DiskSource, HttpSource, 
    ResponseToLines, Transpose
)

Action      = Union[Hashable, HashableDict]
Context     = Union[None, Hashable, HashableDict]

class Interaction:
    """A class to contain all data needed to represent an interaction in a bandit simulation."""

    @overload
    def __init__(self, context: Context, actions: Sequence[Action], *, rewards: Sequence[float], **kwargs: Sequence[Any]) -> None:
        ...
        """Instantiate Interaction.

        Args
            context : Features describing the interaction's context. Will be `None` for multi-armed bandit simulations.
            actions : Features describing available actions in the interaction.
            rewards : The rewards that will be revealed to learners based on the action that is taken.
            **kwargs: Additional information beyond reward that will be recorded in the results of a benchmark.
        """

    @overload
    def __init__(self, context: Context, actions: Sequence[Action], *, reveals: Sequence[Any], **kwargs: Sequence[Any]) -> None:
        ...
        """Instantiate Interaction.

        Args
            context : Features describing the interaction's context. Will be `None` for multi-armed bandit simulations.
            actions : Features describing available actions in the interaction.
            reveals : What will be revealed to learners based on the action that is taken.
            **kwargs: Additional information beyond whit is revealed that will be recorded in the results of a benchmark.
        """
    
    def __init__(self, *args, **kwargs) -> None:

        assert len(args) == 2, "An unexpected number of positional arguments was supplied to Interaction."
        assert kwargs.keys() & {"rewards", "reveals"}, "Interaction requires either a rewards or reveals kwarg."

        context = args[0]
        actions = args[1]
        reveals = kwargs.get("rewards", None) or kwargs["reveals"]
        results = kwargs
        
        assert len(actions) == len(reveals), "Interaction requires information to reveal for each action."

        for singular,plural in [("reward","rewards"), ("reveal","reveals")]:
            if plural in results:
                results[singular] = results[plural]
                del results[plural]

        self._context =  context if not isinstance(context,dict) else HashableDict(context)
        self._actions = [ action if not isinstance(action, dict) else HashableDict(action) for action in actions ]
        self._reveals = reveals
        self._results = results

    def _is_sparse(self, feats):

        if isinstance(feats,dict):
            return True

        if not isinstance(feats, collections.Sequence):
            return False

        if len(feats) != 2:
            return False

        if not isinstance(feats[0], collections.Sequence) or not isinstance(feats[1],collections.Sequence):
            return False

        if len(feats[0]) != len(feats[1]):
            return False

        if isinstance(feats[0],str) or isinstance(feats[1],str):
            return False

        return True

    def _flatten(self, feats):

        if not isinstance(feats, collections.Sequence) or isinstance(feats, (str,dict)):
            return feats

        if not self._is_sparse(feats):

            flattened_dense_values = []

            for val in feats:
                if isinstance(val,(list,tuple,bytes)):
                    flattened_dense_values.extend(val)
                else:
                    flattened_dense_values.append(val)
            
            return tuple(flattened_dense_values)
        else:
            keys = []
            vals = []

            for key,val in zip(*feats):

                if isinstance(val, (list,tuple,bytes)):
                    for sub_key,sub_val in enumerate(val):
                        keys.append(f"{key}_{sub_key}")
                        vals.append(sub_val)
                else:
                    keys.append(key)
                    vals.append(val)

            return HashableDict(zip(keys,vals))

    @property
    def context(self) -> Context:
        """The interaction's context description."""

        return self._flatten(self._context)

    @property
    def actions(self) -> Sequence[Action]:
        """The interaction's available actions."""

        return [ self._flatten(action) for action in self._actions ]

    @property
    def reveals(self) -> Sequence[Any]:
        """The interaction's information revealed to learner's based on the selected action."""
        return self._reveals

    @property
    def results(self) -> Sequence[Dict[str,Any]]:
        """The interaction's information recorded in results based on the selected action."""
        return self._results
    
    def reveal(self, action: Action) -> Any:
        """Return the information revealed by taking the given action."""
        return self._reveals[self._actions.index(action)]

    def result(self, action: Action) -> Dict[str,Any]:
        """Return the recorded result for taking the given action."""
        result = {}
        aindex = self._actions.index(action) 

        for key in self._results:
            result[key] = self._results[key][aindex]

        return result

class Simulation(Source[Iterable[Interaction]], ABC):
    """The simulation interface."""

    @property
    @abstractmethod
    def params(self) -> Dict[str,Any]:
        """Paramaters describing the simulation.

        Remarks:
            These will be simulation columns in coba.benchmark.Result.
        """
        ...
    
    @abstractmethod
    def read(self) -> Iterable[Interaction]:
        """The sequence of interactions in a simulation.

        Remarks:
            This function should always be "re-iterable".
        """
        ...

class MemorySimulation(Simulation):
    """A Simulation implementation created from in memory sequences of contexts, actions and rewards."""

    def __init__(self, interactions: Sequence[Interaction]) -> None:
        """Instantiate a MemorySimulation.

        Args:
            interactions: The sequence of interactions in this simulation.
        """

        self._interactions = interactions

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return {}

    def read(self) -> Iterable[Interaction]:
        """Read the interactions in this simulation."""
        return self._interactions

class ClassificationSimulation(Simulation):
    """A simulation created from classification dataset with features and labels.

    ClassificationSimulation turns labeled observations from a classification data set
    into interactions. For each interaction the feature set becomes the context and 
    all possible labels become the actions. Rewards for each interaction are created by 
    assigning a reward of 1 for taking the correct action (i.e., choosing the correct
    label)) and a reward of 0 for taking any other action (i.e., choosing any of the
    incorrect lables).
    """

    def __init__(self, features: Sequence[Any], labels: Union[Sequence[Action], Sequence[List[Action]]] ) -> None:
        """Instantiate a ClassificationSimulation.

        Args:
            features: The collection of features used for the original classifier problem.
            labels: The collection of labels assigned to each observation of features.
        """

        assert len(features) == len(labels), "Mismatched lengths of features and labels"

        #how can we tell the difference between featurized labels and multilabels????
        #for now we will assume multilables will be passed in as arrays not tuples...
        if not isinstance(labels[0], collections.Hashable):
            labels_flat = list(chain.from_iterable(labels))
        else:
            labels_flat = list(labels)

        reveal        = lambda action,label: int(is_label(action,label) or in_multilabel(action,label)) #type: ignore
        is_label      = lambda action,label: action == label #type: ignore
        in_multilabel = lambda action,label: isinstance(label,collections.Sequence) and action in label #type: ignore

        # shuffling so that action order contains no statistical information
        # sorting so that the shuffled values are always shuffled in the same order
        actions  = CobaRandom(1).shuffle(sorted(set(labels_flat))) 
        contexts = features
        rewards  = [ [ reveal(action,label) for action in actions ] for label in labels ]

        self._interactions = [ Interaction(c,a,rewards=r) for c,a,r in zip(contexts, repeat(actions), rewards) ]

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return { }

    def read(self) -> Iterable[Interaction]:
        """Read the interactions in this simulation."""

        return self._interactions

class RegressionSimulation(Simulation):
    """A simulation created from regression dataset with features and labels.
    
    RegressionSimulation turns labeled observations from a regression data set
    into interactions. For each interaction the feature set becomes the context and 
    all possible labels become the actions. Rewards for each interaction are created by 
    assigning a minus absolute error. Rewards are close to zero for taking actions that are 
    closer to the correct action (label) and lower ones for being far from the correct action.
    
    """

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return {}

    def __init__(self, features: Sequence[Any], labels: Sequence[Number] ) -> None:
        """Instantiate a RegressionSimulation.
    
        Args:
            features: The collection of features used for the original regression problem.
            labels: The collection of labels assigned to each observation of features.
        """

        assert len(features) == len(labels), "Mismatched lengths of features and labels"

        reward   = lambda action,label: 1-abs(float(action)-float(label))
        contexts = features
        actions  = CobaRandom(1).shuffle(sorted(set(labels)))
        rewards  = [ [ reward(action,label) for action in actions ] for label in labels ]

        self._interactions = [ Interaction(c,a,rewards=r) for c,a,r in zip(contexts, repeat(actions), rewards) ]

    def read(self) -> Iterable[Interaction]:
        """Read the interactions in this simulation."""
        return self._interactions

class LambdaSimulation(Simulation):
    """A Simulation created from lambda functions that generate contexts, actions and rewards.

    Remarks:
        This implementation is useful for creating simulations from defined distributions.
    """

    def __init__(self,
        n_interactions: int,
        context       : Callable[[int               ],Context         ],
        actions       : Callable[[int,Context       ],Sequence[Action]], 
        reward        : Callable[[int,Context,Action],float           ]) -> None:
        """Instantiate a LambdaSimulation.

        Args:
            n_interactions: How many interactions the LambdaSimulation should have.
            context: A function that should return a context given an index in `range(n_interactions)`.
            actions: A function that should return all valid actions for a given index and context.
            reward: A function that should return the reward for the index, context and action.
        """

        self._interactions: List[Interaction] = []

        for i in range(n_interactions):
            _context  = context(i)
            _actions  = actions(i, _context)
            _rewards  = [ reward(i, _context, _action) for _action in _actions]

            self._interactions.append(Interaction(_context, _actions, rewards=_rewards))

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return {}

    def read(self) -> Iterable[Interaction]:
        """Read the interactions in this simulation."""
        
        return self._interactions

class ReaderSimulation(Simulation):

    def __init__(self, 
        reader      : Filter[Iterable[str], Any], 
        source      : Union[str,Source[Iterable[str]]], 
        label_column: Union[str,int], 
        with_header : bool=True) -> None:
        
        self._reader = reader

        if isinstance(source, str) and source.startswith('http'):
            self._source = Pipe.join(HttpSource(source), [ResponseToLines()])
        elif isinstance(source, str):
            self._source = DiskSource(source)
        else:
            self._source = source
        
        self._label_column = label_column
        self._with_header  = with_header
        self._interactions = cast(Optional[Sequence[Interaction]], None)

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return {"source": str(self._source) }

    def read(self) -> Iterable[Interaction]:
        """Read the interactions in this simulation."""
        return self._load_interactions()

    def _load_interactions(self) -> Sequence[Interaction]:
        parsed_rows_iter = iter(self._reader.filter(self._source.read()))

        if self._with_header:
            header = next(parsed_rows_iter)
        else:
            header = []

        if isinstance(self._label_column, str):
            label_col_index = header.index(self._label_column)
        else:
            label_col_index = self._label_column

        parsed_cols = list(Transpose().filter(parsed_rows_iter))
        
        label_col    = parsed_cols.pop(label_col_index)
        feature_rows = list(Transpose().filter(parsed_cols))

        is_sparse_labels = len(label_col) == 2 and isinstance(label_col[0],tuple) and isinstance(label_col[1],tuple)
        
        if is_sparse_labels:
            dense_labels: List[Any] = ['0']*len(feature_rows)
            
            for label_row, label_val in zip(*label_col): #type:ignore
                dense_labels[label_row] = label_val

        else:
            dense_labels = list(label_col)

        return ClassificationSimulation(feature_rows, dense_labels).read()

    def __repr__(self) -> str:
        return str(self._source)

class CsvSimulation(ReaderSimulation):
    def __init__(self, source:Union[str,Source[Iterable[str]]], label_column:Union[str,int], with_header:bool=True) -> None:
        super().__init__(CsvReader(), source, label_column, with_header)

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return { "csv": super().params["source"] }

class ArffSimulation(ReaderSimulation):
    def __init__(self, source:Union[str,Source[Iterable[str]]], label_column:Union[str,int]) -> None:
        super().__init__(ArffReader(skip_encoding=[label_column]), source, label_column)

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return { "arff": super().params["source"] }

class LibsvmSimulation(ReaderSimulation):
    def __init__(self, source:Union[str,Source[Iterable[str]]]) -> None:
        super().__init__(LibSvmReader(), source, 0, False)

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return { "libsvm": super().params["source"] }

class ManikSimulation(ReaderSimulation):
    def __init__(self, source:Union[str,Source[Iterable[str]]]) -> None:
        super().__init__(ManikReader(), source, 0, False)

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return { "manik": super().params["source"] }

class ValidationSimulation(LambdaSimulation):
    
    def __init__(self, n_interactions: int=500, n_actions: int=10, n_features: int=10, context_features:bool = True, action_features:bool = True, sparse: bool=False, seed:int=1, make_binary=False) -> None:

        self._n_bandits        = n_actions
        self._n_features       = n_features
        self._context_features = context_features
        self._action_features  = action_features
        self._seed             = seed
        self._make_binary      = make_binary

        r = CobaRandom(seed)

        context: Callable[[int               ], Context         ]
        actions: Callable[[int,Context       ], Sequence[Action]]
        rewards: Callable[[int,Context,Action], float           ]

        sparsify  = lambda x: (tuple(range(len(x))), tuple(x)) if sparse else tuple(x)
        unsparse  = lambda x: x[1] if sparse else x
        normalize = lambda X: [x/sum(X) for x in X]

        if not context_features and not action_features:

            means = [ m/n_actions + 1/(2*n_actions) for m in r.randoms(n_actions) ]

            actions_features = []
            for i in range(n_actions):
                action = [0] * n_actions
                action[i] = 1
                actions_features.append(tuple(action))

            context = lambda i     : None
            actions = lambda i,c   : sparsify(actions_features)
            rewards = lambda i,c,a : means[unsparse(a).index(1)] + (r.random()-.5)/n_actions

        if context_features and not action_features:
            #normalizing allows us to make sure our reward is in [0,1]
            bandit_thetas = [ r.randoms(n_features) for _ in range(n_actions) ]
            theta_totals  = [ sum(theta) for theta in bandit_thetas]
            bandit_thetas = [ [t/norm for t in theta ] for theta,norm in zip(bandit_thetas,theta_totals)]
            scales        = [ 1 for _ in range(n_features) ]

            actions_features = []
            for i in range(n_actions):
                action = [0] * n_actions
                action[i] = 1
                actions_features.append(tuple(action))

            context = lambda i     : sparsify([ f*s for f,s in zip(r.randoms(n_features),scales) ])
            actions = lambda i,c   : [sparsify(af) for af in actions_features]
            rewards = lambda i,c,a : sum([cc*t for cc,t in zip(unsparse(c),bandit_thetas[unsparse(a).index(1)])])

        if not context_features and action_features:

            theta = r.randoms(n_features)

            context = lambda i     :   None
            actions = lambda i,c   : [ sparsify(normalize(r.randoms(n_features))) for _ in range(n_actions) ]
            rewards = lambda i,c,a : float(sum([cc*t for cc,t in zip(theta,unsparse(a))]))

        if context_features and action_features:

            context = lambda i     :   sparsify(r.randoms(n_features))
            actions = lambda i,c   : [ sparsify(normalize(r.randoms(n_features))) for _ in range(n_actions) ]
            rewards = lambda i,c,a : sum([cc*t for cc,t in zip(unsparse(c),unsparse(a))])

        super().__init__(n_interactions, context, actions, rewards)

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return { }

    def read(self) -> Iterable[Interaction]:
        for i in super().read():
            if self._make_binary:
                yield Interaction(i.context, i.actions, rewards = [ int(r == max(i.reveals)) for r in i.reveals ] )
            else:
                yield i

    def __repr__(self) -> str:
        return "ValidationSimulation"



import json

from math import isnan
from itertools import compress
from hashlib import md5
from numbers import Number
from typing import Optional, Tuple, Sequence, Any, List, Iterable, Dict

from coba.pipes import Source, HttpSource
from coba.config import CobaConfig, CobaException

from coba.simulations.core import Context, Action, Interaction, Simulation, ClassificationSimulation, RegressionSimulation

class OpenmlSource(Source[Tuple[Sequence[Context], Sequence[Action]]]):

    def __init__(self, id:int, problem_type:str = "classification", nominal_as_str:bool=False, md5_checksum:str = None):
        
        assert problem_type in ["classification", "regression"]
        
        self._data_id        = id
        self._md5_checksum   = md5_checksum
        self._problem_type   = problem_type 
        self._nominal_as_str = nominal_as_str
        self._cached_urls    = []

        #add flag to allow regression simulations
        #if a known exception happens still cache the query results

    def read(self) -> Tuple[Sequence[Sequence[Any]], Sequence[Any]]:
        
        #placing some of these at the top would cause circular references
        from coba.encodings import Encoder, NumericEncoder, OneHotEncoder, StringEncoder
        from coba.pipes     import Encode, Transpose

        try:
            data_id        = self._data_id
            md5_checksum   = self._md5_checksum

            dataset_description  = self._get_dataset_description(data_id)

            if dataset_description['status'] == 'deactivated':
                raise CobaException(f"Openml {data_id} has been deactivated. This is often due to flags on the data.")
            
            feature_descriptions = self._get_feature_descriptions(data_id)

            headers : List[str]     = []
            encoders: List[Encoder] = []
            ignored : List[bool]    = []
            target  : str           = ""

            for tipe in feature_descriptions:

                headers.append(tipe['name'].lower())
                ignored.append(tipe['is_ignore'] == 'true' or tipe['is_row_identifier'] == 'true')

                if tipe['is_target'] == 'true':
                    target = tipe['name'].lower()
                    
                if tipe['data_type'] == 'numeric':
                    encoders.append(NumericEncoder())  
                elif tipe['data_type'] == 'nominal':
                    if self._nominal_as_str:
                        encoders.append(StringEncoder())
                    else:
                        encoders.append(OneHotEncoder(singular_if_binary=True))
                else:
                    ignored[-1] = True
                    encoders.append(StringEncoder())

            if target != "":
                target_encoder = encoders[headers.index(target)]
                problem_encoder = NumericEncoder if self._problem_type == "regression" else (OneHotEncoder,StringEncoder)

            if target == "" or not isinstance(target_encoder, problem_encoder):
                target = self._get_target_for_problem_type(data_id)
                ignored[headers.index(target)] = False

            if self._problem_type == "classification":
                target_encoder = StringEncoder() if self._nominal_as_str else OneHotEncoder(singular_if_binary=False)
                encoders[headers.index(target)] = target_encoder

            file_rows = self._get_dataset_rows(dataset_description["file_id"], target, md5_checksum)
            is_sparse = isinstance(file_rows[0], tuple) and len(file_rows[0]) == 2

            if is_sparse:
                file_headers  = [ header.lower() for header in file_rows.pop(0)[1]]
            else:
                file_headers  = [ header.lower() for header in file_rows.pop(0)]

            file_cols = list(Transpose().filter(file_rows))

            for ignored_header in compress(headers, ignored):
                if ignored_header in file_headers:
                    file_cols.pop(file_headers.index(ignored_header))
                    file_headers.remove(ignored_header)

            file_encoders = [ encoders[headers.index(file_header)] for file_header in file_headers]

            if is_sparse:
                label_col = file_cols[file_headers.index(target)]
                
                dense_label_col = ['0']*len(file_rows)
                
                for index, value in zip(label_col[0], label_col[1]):
                    dense_label_col[index] = value

                file_cols[file_headers.index(target)] = (tuple(range(len(file_rows))), tuple(dense_label_col))

            file_cols    = list(Encode(file_encoders).filter(file_cols))
            label_col    = file_cols.pop(file_headers.index(target))
            feature_rows = list(Transpose().filter(file_cols))

            if is_sparse:
                label_col = label_col[1]

            no_missing_values = [ not any(isinstance(val,Number) and isnan(val) for val in row) for row in feature_rows ]

            if not any(no_missing_values):
                raise CobaException(f"Every example in openml {data_id} has a missing value. The simulation is being ignored.")

            feature_rows = list(compress(feature_rows, no_missing_values))
            label_col    = list(compress(label_col, no_missing_values)) 
            
            return feature_rows, label_col

        except KeyboardInterrupt:
            #we don't want to clear the cache in the case of a KeyboardInterrupt
            raise
        
        except CobaException:
            raise

        except Exception:
            #if something unexpected went wrong clear the
            #cache just in case it was corrupted somehow
            
            for key in self._cached_urls:
                CobaConfig.Cacher.rmv(key)
            
            raise

    def _get_url(self, url:str, checksum:str=None) -> bytes:
        
        if url in CobaConfig.Cacher:
            bites = CobaConfig.Cacher.get(url)
        else:

            api_key  = CobaConfig.Api_Keys['openml']
            response = HttpSource(url + (f'?api_key={api_key}' if api_key else '')).read()

            if response.status_code == 412:
                if 'please provide api key' in response.text:
                    message = (
                        "An API Key is needed to access openml's rest API. A key can be obtained by creating an "
                        "openml account at openml.org. Once a key has been obtained it should be placed within "
                        "~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")
                    raise CobaException(message) from None

                if 'authentication failed' in response.text:
                    message = (
                        "The API Key you provided no longer seems to be valid. You may need to create a new one"
                        "longing into your openml account and regenerating a key. After regenerating the new key "
                        "should be placed in ~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")
                    raise CobaException(message) from None

            if response.status_code == 404:
                message = (
                    "We're sorry but we were unable to find the requested dataset on openml. The most likely cause "
                    "for this is openml not providing the requested dataset in a format that COBA can process.")
                raise CobaException(message) from None

            if "Usually due to high server load" in response.text:
                message = (
                    "Openml has experienced an error that they believe is the result of high server loads."
                    "Openml recommends that you try again in a few seconds. Additionally, if not already "
                    "done, consider setting up a DiskCache in coba config to reduce the number of openml "
                    "calls in the future.")
                raise CobaException(message) from None

            if '' == response.text:
                raise CobaException("Openml experienced an unexpected error. Please try requesting the data again.") from None

            bites = response.content

            if url not in CobaConfig.Cacher:
                self._cached_urls.append(url)
                CobaConfig.Cacher.put(url,bites)

        if checksum is not None and md5(bites).hexdigest() != checksum:
            
            #if the cache has become corrupted we need to clear it
            CobaConfig.Cacher.rmv(url)

            message = (
                f"The response from {url} did not match the given checksum {checksum}. This could be the result "
                "of network errors or the file becoming corrupted. Please consider downloading the file again. "
                "If the error persists you may want to manually download and reference the file.")
            raise CobaException(message) from None

        return bites.decode('utf-8')

    def _get_dataset_description(self, data_id:int) -> Dict[str,Any]:

        description_txt = self._get_url(f'https://www.openml.org/api/v1/json/data/{data_id}')
        description_obj = json.loads(description_txt)["data_set_description"]

        return description_obj

    def _get_feature_descriptions(self, data_id:int) -> Sequence[Dict[str,Any]]:
        
        types_txt = self._get_url(f'https://www.openml.org/api/v1/json/data/features/{data_id}')
        types_obj = json.loads(types_txt)["data_features"]["feature"]

        return types_obj

    def _get_dataset_rows(self, file_id:str, target:str, md5_checksum:str) -> Any:
            from coba.pipes import ArffReader, CsvReader, Encode, Transpose
            
            csv_url  = f"http://www.openml.org/data/v1/get_csv/{file_id}"
            arff_url = f"http://www.openml.org/data/v1/download/{file_id}"

            if arff_url in CobaConfig.Cacher:
                text = self._get_url(arff_url, md5_checksum)
                return list(ArffReader(skip_encoding=[target]).filter(text.splitlines()))

            try:
                text = self._get_url(csv_url, md5_checksum)
                return list(CsvReader().filter(text.splitlines()))
            
            except:
                text = self._get_url(arff_url, md5_checksum)
                return list(ArffReader(skip_encoding=[target]).filter(text.splitlines()))

    def _get_target_for_problem_type(self, data_id:int):

        text  = self._get_url(f'https://www.openml.org/api/v1/json/task/list/data_id/{data_id}')
        tasks = json.loads(text).get("tasks",{}).get("task",[])

        task_type = 1 if self._problem_type == "classification" else 2

        for task in tasks:
            if task["task_type_id"] == task_type: #aka, classification task
                for input in task['input']:
                    if input['name'] == 'target_feature':
                        return input['value'] #just take the first one

        raise CobaException(f"Openml {data_id} does not appear to be a {self._problem_type} dataset")

class OpenmlSimulation(Simulation):
    """A simulation created from openml data with features and labels.

    OpenmlSimulation turns labeled observations from a classification data set,
    into interactions. For each interaction the feature set becomes the context and 
    all possible labels become the actions. Rewards for each interaction are created by 
    assigning a reward of 1 for taking the correct action (i.e., choosing the correct
    label) and a reward of 0 for taking any other action (i.e., choosing any of the
    incorrect lables).
    """

    def __init__(self, id: int, simulation_type:str = "classification", nominal_as_str:bool = False, md5_checksum: str = None) -> None:
        self._source = OpenmlSource(id, simulation_type, nominal_as_str, md5_checksum)
        self._interactions: Optional[Sequence[Interaction]] = None

    @property
    def params(self) -> Dict[str, Any]:
        """Paramaters describing the simulation."""
        return { "openml": self._source._data_id}

    def read(self) -> Iterable[Interaction]:
        """Read the interactions in this simulation."""

        features,labels = self._source.read()

        if isinstance(labels[0],Number):
            return RegressionSimulation(features,labels).read()
        else:
            return ClassificationSimulation(features,labels).read()

    def __repr__(self) -> str:
        return f'{{"OpenmlSimulation":{self._source._data_id}}}'



"""The vowpal module contains classes to make it easier to interact with pyvw."""

import re
import collections

from itertools import repeat
from numbers import Number
from os import devnull
from typing import Any, Dict, Union, Sequence, overload, cast, Optional, Tuple

from coba.config import CobaException
from coba.utilities import PackageChecker, redirect_stderr
from coba.simulations import Context, Action

from coba.learners.core import Learner, Probs, Info

Feature  = Union[str,Number]
Features = Union[Dict[str,Feature], Sequence[Feature], Feature]

class VowpalMediator:
    
    @staticmethod
    def prep_features(features: Features) -> Sequence[Tuple[str,float]]:
        if not features:
            return []
        elif isinstance(features, dict):
            return [ t[1] if isinstance(t[1],str) else t for t in features.items() if t[1] != 0 ]
        elif isinstance(features, str):
            return [features]
        elif isinstance(features, Number):
            return [(0,features)]
        elif isinstance(features, collections.Sequence):
            if isinstance(features[0], collections.Sequence) and len(features[0]) == 2:
                return features
            else:
                return [ t[1] if isinstance(t[1],str) else t for t in enumerate(features) if t[1] != 0 ]
        else:
            raise Exception("Unrecognized features passed to VowpalLearner.")

    @staticmethod
    def make_learner(args:str):
        PackageChecker.vowpalwabbit('VVW.make_learner')
        from vowpalwabbit import pyvw

        return pyvw.vw(args)
    
    @staticmethod
    def make_example(vw, ns:Dict[str,Any], label:Optional[str], label_type:int):
        PackageChecker.vowpalwabbit('VW.make_example')
        from vowpalwabbit.pyvw import example

        ex = example(vw, ns, label_type)
        if label: ex.set_label_string(label)
        ex.setup_example()
        
        return ex

class VowpalLearner(Learner):
    """A learner using Vowpal Wabbit's contextual bandit command line interface.

    Remarks:
        This learner requires that the Vowpal Wabbit package be installed. This package can be
        installed via `pip install vowpalwabbit`. To learn more about solving contextual bandit
        problems with Vowpal Wabbit see https://vowpalwabbit.org/tutorials/contextual_bandits.html
        and https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms.
    """

    @overload
    def __init__(self, *, epsilon: float = 0.1, adf: bool = True, seed: Optional[int] = 1) -> None:
        """Instantiate a VowpalLearner.
        Args:
            epsilon: A value between 0 and 1. If provided, exploration will follow epsilon-greedy.
            adf: Indicate whether cb_explore or cb_explore_adf should be used.
            seed: The seed used by VW to generate any necessary random numbers.
        """
        ...

    @overload
    def __init__(self, *, bag: int, adf: bool = True, seed: Optional[int] = 1) -> None:
        """Instantiate a VowpalLearner.
        Args:
            bag: An integer value greater than 0. This value determines how many separate policies will be
                learned. Each policy will be learned from bootstrap aggregation, making each policy unique. 
                When predicting one policy will be selected according to a uniform distribution and followed.
            adf: Indicate whether cb_explore or cb_explore_adf should be used.
            seed: The seed used by VW to generate any necessary random numbers.
        """
        ...

    @overload
    def __init__(self, *, cover: int, seed: Optional[int] = 1) -> None:
        """Instantiate a VowpalLearner.
        Args:
            cover: An integer value greater than 0. This value value determines how many separate policies will be
                learned. These policies are learned in such a way to explicitly optimize policy diversity in order
                to control exploration. When predicting one policy will be selected according to a uniform distribution
                and followed. For more information on this algorithm see Agarwal et al. (2014).
            seed: The seed used by VW to generate any necessary random numbers.
        
        References:
            Agarwal, Alekh, Daniel Hsu, Satyen Kale, John Langford, Lihong Li, and Robert Schapire. "Taming 
            the monster: A fast and simple algorithm for contextual bandits." In International Conference on 
            Machine Learning, pp. 1638-1646. 2014.
        """
        ...

    @overload
    def __init__(self, *, softmax: float, seed: Optional[int] = 1) -> None:
        """Instantiate a VowpalLearner.
        Args:
            softmax: An exploration parameter with 0 indicating uniform exploration is desired and infinity
                indicating that no exploration is desired (aka, greedy action selection only). For more info
                see `lambda` at https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms.
            seed: The seed used by VW to generate any necessary random numbers.
        """
        ...

    @overload
    def __init__(self, args:str) -> None:
        ...
        """Instantiate a VowpalLearner.
        Args:
            args: Command line argument to instantiates a Vowpal Wabbit contextual bandit learner. 
                For examples and documentation on how to instantiate VW learners from command line arguments see 
                https://github.com/VowpalWabbit/vowpal_wabbit/wiki/Contextual-Bandit-algorithms. It is assumed that
                either the --cb_explore or --cb_explore_adf flag is used. When formatting examples for VW, context
                features are namespaced with `s` and action features, when relevant, are namespaced with `a`.
        """

    def __init__(self, *args, **kwargs) -> None:
        """Instantiate a VowpalLearner with the requested VW learner and exploration."""

        interactions = "--interactions ssa --interactions sa --ignore_linear s"

        if not args and 'seed' not in kwargs:
            kwargs['seed'] = 1

        if not args and all(e not in kwargs for e in ['epsilon', 'softmax', 'bag', 'cover']): 
            kwargs['epsilon'] = 0.1

        if len(args) > 0:

            assert "--cb" in args[0], "VowpalLearner was instantiated without a cb learner being defined."

            self._adf  = "--cb_explore_adf" in args[0]
            self._args = cast(str,args[0])

            self._args = re.sub("--cb_explore_adf\s+", '', self._args, count=1)
            self._args = re.sub("--cb_explore(\s+\d+)?\s+", '', self._args, count=1)

        elif 'epsilon' in kwargs:
            self._adf  = kwargs.get('adf', True)
            self._args = interactions + f" --epsilon {kwargs['epsilon']}"

        elif 'softmax' in kwargs:
            self._adf  = True
            self._args = interactions + f" --softmax --lambda {kwargs['softmax']}"

        elif 'bag' in kwargs:
            self._adf  = kwargs.get('adf',True)
            self._args = interactions + f" --bag {kwargs['bag']}"

        elif 'cover' in kwargs:
            self._adf  = False
            self._args = interactions + f" --cover {kwargs['cover']}"

        if 'seed' in kwargs and kwargs['seed'] is not None:
            self._args += f" --random_seed {kwargs['seed']}"

        self._precision                 = kwargs.get('precision',5) 
        self._actions: Sequence[Action] = []
        self._vw                        = None

    @property
    def params(self) -> Dict[str, Any]:
        """The parameters of the learner.

        See the base class for more information
        """

        return {"family": "vw", 'args': self._cli_args([None])}

    def predict(self, context: Context, actions: Sequence[Action]) -> Tuple[Probs, Info]:
        """Determine a PMF with which to select the given actions.

        Args:
            context: The context we're currently in. See the base class for more information.
            actions: The actions to choose from. See the base class for more information.

        Returns:
            The probability of taking each action. See the base class for more information.
        """

        if self._vw is None:
             #type: ignore
            # vowpal has an annoying warning that is written to stderr whether or not we provide
            # the --quiet flag. Therefore, we temporarily redirect all stderr output to null so that
            # this warning isn't shown during creation. It should be noted this isn't thread-safe
            # so if you are here because of strange problems with threads we may just need to suck
            # it up and accept that there will be an obnoxious warning message.
            with open(devnull, 'w') as f, redirect_stderr(f):
                self._vw = VowpalMediator.make_learner(self._cli_args(actions) + " --quiet")

            if not self._adf:
                self._actions = actions

        if not self._adf and (len(actions) != len(self._actions) or any(a not in self._actions for a in actions)):
            raise CobaException("Actions are only allowed to change between predictions with `--cb_explore_adf`.")

        info = (actions if self._adf else self._actions)

        shared  = self._shared(context)
        adfs    = self._adfs(actions)

        if self._adf:
            probs = self._vw.predict(self._examples(shared,adfs))
        else:
            probs = self._vw.predict(self._example(shared))

        return probs, info

    def learn(self, context: Context, action: Action, reward: float, probability: float, info: Info) -> None:
        """Learn from the given interaction.

        Args:
            key: The key identifying the interaction this observed reward came from.
            context: The context we're learning about. See the base class for more information.
            action: The action that was selected in the context. See the base class for more information.
            reward: The reward that was gained from the action. See the base class for more information.
            probability: The probability that the given action was taken.
        """

        assert self._vw is not None, "You must call predict before learn in order to initialize the VW learner"

        actions = info
        shared  = self._shared(context)
        adfs    = self._adfs(actions)
        labels  = self._labels(actions, action, reward, probability)
        label   = labels[actions.index(action)]

        if self._adf:
            self._vw.learn(self._examples(shared, adfs, labels))
        else:
            self._vw.learn(self._example(shared, label))

    def _cli_args(self, actions: Optional[Sequence[Action]]) -> str:

        if self._adf:
            return "--cb_explore_adf" + " " + self._args
        else:
            return f"--cb_explore {len(actions) if actions else ''}" + " " + self._args

    def _examples(self, shared: Dict[str,Any], adfs: Sequence[Dict[str,Any]] = None, labels: Sequence[str] = None):

        shared = { ns:VowpalMediator.prep_features(v) for ns,v in shared.items() }
        examples = []
        for adf,label in zip(adfs, labels or repeat(None)):
            adf = { ns:VowpalMediator.prep_features(v) for ns,v in adf.items() }
            examples.append(self._example({**shared,**adf}, label))
        
        return examples

    def _example(self, ns, label=None):
        return VowpalMediator.make_example(self._vw, ns, label, 4)

    def _labels(self,actions,action,reward:float,prob:float) -> Sequence[Optional[str]]:
        return [ f"{i+1}:{round(1-reward,5)}:{round(prob,5)}" if a == action else None for i,a in enumerate(actions)]

    def _shared(self,context) -> Dict[str,Any]:
        return {} if not context else { 's': VowpalMediator.prep_features(context) }

    def _adfs(self,actions) -> Sequence[Dict[str,Any]]:
        return [ {'a': VowpalMediator.prep_features(a)} for a in actions]
